#define STB_IMAGE_IMPLEMENTATION
#define STBI_STATIC
#define STBI_ONLY_PNG
#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"
#endif
#include "stb_image.h"
#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif
#include <stdio.h>

typedef unsigned char  u8;
typedef unsigned short u16;
typedef unsigned int   u32;

#ifndef __GNUC__
static u32 __builtin_bswap32(u32 x) {
	u32 a = x & 0xFF;
	u32 b = (x >> 8) & 0xFF;
	u32 c = (x >> 16) & 0xFF;
	u32 d = (x >> 24) & 0xFF;
	return (a << 24) | (b << 16) | (c << 8) | d;
}
#endif

static void place(u8 *spritesheet, int spritesheet_width, char const *filename, int x1, int y1) {
	int channels = 0, w = 0, h = 0;
	int y;
	size_t row_size_in_bytes;
	u8 *data = stbi_load(filename, &w, &h, &channels, 4);
	if (!data) {
		fprintf(stderr, "Couldn't read image: %s\n", filename);
		exit(-1);
	}
	if (channels != 4) {
		fprintf(stderr, "Invalid number of channels: %d\n", channels);
		exit(-1);
	}
	row_size_in_bytes = (size_t)w * 4;

	for (y = 0; y < h; ++y) {
		u8 *in = &data[(y*w) * 4]; /* in = (0, y) */
		u8 *out = &spritesheet[((y+y1)*spritesheet_width + x1) * 4]; /* out = (x1, y+y1) */
		u32 *row_start = (u32 *)out, *row_end = (u32 *)(out + row_size_in_bytes);
		u32 *p;

		for (p = row_start; p != row_end; ++p) {
			if (*p) {
				fprintf(stderr, "Overlap @ (%ld, %d).\n", (long)(p - (u32 *)out), y);
				exit(-1);
			}
		}
		memcpy(out, in, row_size_in_bytes); /* copy one row */
		for (p = row_start; p != row_end; ++p) {
			/* rearrange bytes to be in the right format */
			*p = (__builtin_bswap32(*p) >> 8) | (*p & 0xFF000000);
		}
	}

	stbi_image_free(data);
}

static void write_u32(FILE *f, u32 u) {
	fwrite(&u, 1, sizeof u, f);
}
static void write_u16(FILE *f, u16 u) {
	fwrite(&u, 1, sizeof u, f);
}

#ifdef _MSC_VER
#pragma pack(1)
#endif
typedef struct {
	char B, M;
	u32 file_size;
	u32 reserved;
	u32 data_offset;
	u32 header_size;
	u32 width, height;
	u16 planes;
	u16 bits_per_pixel;
	u32 compression;
	u32 image_size;
	u32 horiz_ppm, vert_ppm;
	u32 clr_used, clr_important;
	u32 red_mask, green_mask, blue_mask, alpha_mask;
	u32 cs_type;
	u32 end1r, end1g, end1b, end2r, end2g, end2b, end3r, end3g, end3b;
	u32 gamma_red, gamma_green, gamma_blue;
} 
#ifdef __GNUC__
__attribute__((packed)) 
#endif
BitmapHeader;

int main(void) {
	int width = 600, height = 600;
	u32 header_size = sizeof(BitmapHeader);
	u32 data_offset = header_size;
	u32 bytes_of_data = width * (u32)height * 4;
	u8 *spritesheet = calloc(width*height,4);
	FILE *out = NULL;
	

	#define img(name, x, y) place(spritesheet, width, name, x, y)
	img("pion.png", 20, 20); /* size: 256x256 */
	img("arrow.png", 20, 290); /* size: 250x100 */
	img("attack.png", 280, 290); /* size: 250x100 */
	img("cbox.png", 300, 20); /* size: 128x128 */
	img("arrowhead.png", 300, 160); /* size: 128x128 */
	img("cbox_checked.png", 450, 20); /* size: 128x128 */
	#undef img

	out = fopen("spritesheet.bmp", "wb");
	
	{
		BitmapHeader h = {0};
		h.B = 'B'; h.M = 'M';
		h.file_size = bytes_of_data + header_size;
		h.reserved = 0;
		h.data_offset = data_offset;
		h.header_size = header_size - offsetof(BitmapHeader, header_size); /* subtract things which technically aren't part of this header */
		h.width = (u32)width;
		h.height = (u32)height;
		h.planes = 1;
		h.bits_per_pixel = 32;
		h.compression = 3; /* bitfields (this lets us use alpha) */
		h.image_size = bytes_of_data;
		h.horiz_ppm = h.vert_ppm = 11811; /* this is what gimp uses */
		h.clr_used = 0;
		h.clr_important = 0;
		h.red_mask   = 0x00ff0000;
		h.green_mask = 0x0000ff00;
		h.blue_mask  = 0x000000ff;
		h.alpha_mask = 0xff000000;
		h.cs_type = 0x73524742;

		/* we can leave everything else as 0 */
		fwrite(&h, sizeof h, 1, out);
	}
	

	fwrite(spritesheet, bytes_of_data, 1, out);

	fclose(out);
	free(spritesheet);
	return 0;
}
