// @TODO(eventually): nice buttons for editing 

static void editor_default_game(Game *game) {
	u32 w = 5;
	u32 h = 6;
	u32 i;
	Pion *m = game->pions;
	memset(game, 0, sizeof *game);
	game->width = (u16)w;
	game->height = (u16)h;
	game->npions = 2*(h-2);
	for (i = 0; i < w*h; ++i) {
		game->tiles[i] = TL_REG;
	}
	for (i = 1; i < h-1; ++i, ++m) {
		m->player = 0;
		m->pos = hex_pos(0, i);
	}
	for (i = 1; i < h-1; ++i, ++m) {
		m->player = 1;
		m->pos = hex_pos(w-1, i);
	}

	game->nplayers = 2;
	game->starting_points = DFL_STARTING_POINTS;
	game->starting_points_increment = DFL_POINTS_INC;
	start_turn(game);
}

/* 
resize the game's hex grid, e.g.
	resize_game(game, -1, 0, 0, 0) => remove the leftmost column of the grid
	resize_game(game, 0, 0, 1, 1) => add one row above and one below the grid
*/
static void resize_game(State *state, int dleft, int dright, int ddown, int dup) {
	Game *game = &state->game;
	int w = game->width, h = game->height;
	int x, y;
	int new_w = w + dleft + dright, new_h = h + dup + ddown;
	size_t new_area = (size_t)new_w * (size_t)new_h, i;
	u32 mark = tmp_push(state);
	u8 *old_tiles = game->tiles, *new_tiles;
	
	if (new_w <= 0 || new_h <= 0 || new_w > GRID_WIDTH_MAX || new_h > GRID_HEIGHT_MAX)
		return;
	
	new_tiles = tmp_alloc(state, new_area);
	// make newly added tiles TL_REGs 
	for (i = 0; i < new_area; ++i)
		new_tiles[i] = TL_REG;
	
	// copy over tiles to new grid 
	for (y = 0; y < h; ++y) {
		for (x = 0; x < w; ++x) {
			int new_x = x + dleft, new_y = y + ddown;
			if (new_x >= 0 && new_y >= 0 && new_x < new_w && new_y < new_h) {
				new_tiles[new_y*new_w + new_x] = old_tiles[y*w + x];
			}
		}
	}

	{
		Pion *pion, *end;
		for (pion = game->pions, end = pion + game->npions; pion != end; ++pion) {
			HexPos *pos = &pion->pos;
			int new_x = (int)pos->x + dleft, new_y = (int)pos->y + ddown;
			if (new_x >= 0 && new_y >= 0 && new_x < new_w && new_y < new_h) {
				// move pion if necessary 
				pos->x = (u16)new_x;
				pos->y = (u16)new_y;
			} else {
				// remove this pion (we shrunk the grid so this pion isn't on it anymore) 
				if (--game->npions)
					*pion-- = *--end;
			}
		}
	}
	game->width = (u16)new_w;
	game->height = (u16)new_h;
	memcpy(game->tiles, new_tiles, new_area);

	tmp_pop(state, mark);
}

static Pion editor_pion(Editor *editor, HexPos pos) {
	Pion pion = {0};
	pion.player = editor->pion.player;
	pion.pos = pos;
	return pion;
}

/* 
	handle the mouse at (mouse_x, mouse_y) with mask mouse_mask.
	you don't need to call this if mouse_mask is 0.
*/
static void editor_handle_mouse(State *state, int mouse_x, int mouse_y, uint mouse_mask) {
	Game *game = &state->game;
	u8 *tiles = game->tiles;
	Editor *editor = &state->editor;

	HexPos mpos = hex_at_pos(game, pixels_to_gl_coords(state, mouse_x, mouse_y));

	bool mbutton_left  = (mouse_mask & MBUTTON_MASK_LEFT)  != 0;
	bool mbutton_right = (mouse_mask & MBUTTON_MASK_RIGHT) != 0;

	if (hex_in_bounds(game, mpos)) {
		u32 idx = hex_idx(game, mpos);
		switch (editor->placing) {
		case ENT_NONE: break;
		case ENT_TILE: {
			if (mbutton_left) {
				if (tiles[idx] != editor->tile.kind) {
					editor->unsaved_changes = true;
				}
				tiles[idx] = editor->tile.kind;
			} else if (mbutton_right) {
				if (tiles[idx] != TL_NONE) {
					editor->unsaved_changes = true;
				}
				pion_remove_at_pos(game, mpos); // don't allow pion with no tile under it 
				tiles[idx] = TL_NONE;
			}
		} break;
		case ENT_PION: {
			if (mbutton_left) {
				if (tiles[idx] != TL_NONE) {
					editor->unsaved_changes = true;
					// if there's a pion here, remove it first (don't allow 2 pions with same pos) 
					pion_remove_at_pos(game, mpos);

					if (game->npions < PIONS_MAX) {
						Pion *pion = &game->pions[game->npions++];
						*pion = editor_pion(editor, mpos);
					}
				}
			} else if (mbutton_right) {
				if (pion_with_pos(game, mpos)) editor->unsaved_changes = true;
				pion_remove_at_pos(game, mpos);
			}
		} break;
		}
	}
}

static void editor_update_inputs(State *state) {
	Editor *editor = &state->editor;
	Game *game = &state->game;
	// update inputs' text
	sprintf(editor->starting_points_input.text, "%u", game->starting_points);
	sprintf(editor->points_inc_input.text, "%d", game->starting_points_increment);
}

static void editor_init(State *state) {
	Editor *editor = &state->editor;
	memset(editor, 0, sizeof *editor);
	editor_default_game(&state->game);
}

static void editor_save_as(State *state) {
	Editor *editor = &state->editor;
	editor->dialog = EDIALOG_SAVE_AS;
	editor->filename_input.focus = true;
	text_input_clear(&editor->filename_input);
}

static void editor_save(State *state) {
	Editor *editor = &state->editor;
	Game *game = &state->game;
	PlatformProcs *procs = &state->procs;
	char const *map_name = editor->map_name;
	{
		TextInput *start = &editor->starting_points_input, *inc = &editor->points_inc_input;
		i32 start_points, points_inc;
		start->focus = false; inc->focus = false;
		text_input_validate_int(start, 0, STARTING_POINTS_MAX, DFL_STARTING_POINTS);
		text_input_validate_int(inc, -STARTING_POINTS_MAX, +STARTING_POINTS_MAX, DFL_POINTS_INC);

		start_points = str_to_i32(start->text, NULL);
		points_inc   = str_to_i32(inc->text,   NULL);
		
		game->starting_points = (u16)start_points;
		game->starting_points_increment = (i16)points_inc;
	}
	if (!streq(map_name, "")) {
		char filename[512] = "maps/";
		strncat(filename, map_name, 500);
		
		// ensure that the maps directory exists 
		if (procs->mkdir("maps") != -1) {
			FILE *fp = fopen(filename, "wb");
			u32 mark = tmp_push(state);
			u8 *data = tmp_alloc(state, COMM_MAX_BYTES);
			WriteBuffer buffer = mkwbuffer(data, COMM_MAX_BYTES);

			if (fp) {
				buf_write_game_setup(&buffer, game, 0);
				{
					size_t n = buf_bytes_written(&buffer);
					if (fwrite(data, 1, n, fp) == n) {
						editor->unsaved_changes = false;
					} else {
						display_error(state, SCR_EDITOR, false, "Couldn't write to file %s.", filename);
					}
				}

				fclose(fp);
			} else {
				display_error(state, SCR_EDITOR, false, "Couldn't open %s.", filename);
			}
			tmp_pop(state, mark);
		} else {
			display_error(state, SCR_EDITOR, false, "Couldn't create the maps folder.");
		}
	} else {
		editor_save_as(state);
	}
}

static int can_char_be_used_in_map_filename(int c) {
	// @TODO(eventually): utf-8 
	return c == ' ' || c == '_' || c == '-' || isalnum(c);
}

static void editor_open(State *state, char const *map) {
	Editor *editor = &state->editor;
	int ret = map_load(state, map, true, SCR_EDITOR);
	if (ret == 1) {
		// all good 
		editor->dialog = EDIALOG_NONE;
		strcpy(editor->map_name, map);
	} else if (ret == 0) {
		display_error(state, SCR_EDITOR, false, "Map %.16s does not exist.", map);
	} else {
		// oh no 
	}
	
	editor_update_inputs(state);
}

static void editor_new(State *state) {
	Game *game = &state->game;
	Editor *editor = &state->editor;
	editor_default_game(game);
	editor->map_name[0] = '\0';
	editor->unsaved_changes = false;
	editor_update_inputs(state);
}

// run the game editor (the screen where you set up the game) 
static void run_game_editor(Frame *frame) {
	State *state = frame_state(frame);
	Input *input = &frame->input;
	Game *game = &state->game;
	Editor *editor = &state->editor;
	
	Event *events = input->events;
	u32 nevents = input->nevents;
	Event *events_end = events + nevents;
	
	float width_gl = state->width_gl, center_x = width_gl * 0.5f;
	v2 mouse_pos = state->mouse_pos;
	
	bool control_key_down = state->control_key_down;
	bool shift_key_down = state->shift_key_down;
	
	v2 grid_min_coords = V2(0.1f, 0.14f), grid_max_coords = V2(width_gl-0.15f, 0.85f);
	TextFont *big_font = &state->big_font, *main_font = &state->main_font;

	Button arrow_buttons[8] = {0};
	Button *expand_left = &arrow_buttons[0], *expand_right = &arrow_buttons[1],
		*expand_up = &arrow_buttons[2], *expand_down = &arrow_buttons[3], *shrink_left = &arrow_buttons[4], *shrink_right = &arrow_buttons[5],
		*shrink_up = &arrow_buttons[6], *shrink_down = &arrow_buttons[7];

	Button button_yes = {0}, button_no = {0};
	int i;

	for (i = 0; i < 8; ++i) {
		arrow_buttons[i].sprite = SPR_ARROWHEAD; // set all sprites to SPR_ARROWHEAD 
		arrow_buttons[i].size = V2(0.05f, 0.05f);
		arrow_buttons[i].center = true;
		arrow_buttons[i].color = i > 3 ? V4(1,0.5f,0.5f,1) : V4(0.5f,1,0.5f,1);
	}

	expand_left->direction = DIR_LEFT; shrink_left->direction = DIR_RIGHT;
	expand_right->direction = DIR_RIGHT; shrink_right->direction = DIR_LEFT;
	expand_up->direction = DIR_UP; shrink_up->direction = DIR_DOWN;
	expand_down->direction = DIR_DOWN; shrink_down->direction = DIR_UP;

	{
		v2 left_pos = V2(0.05f, 0.5f);
		v2 down_pos = V2(center_x, 0.1f);
		v2 right_pos = V2(width_gl-0.05f, 0.5f);
		v2 up_pos = V2(center_x, 0.9f);
		float dx = 0.03f;

		expand_left->pos = v2_add(left_pos, V2(0, -dx));
		expand_right->pos = v2_add(right_pos, V2(0, -dx));
		expand_up->pos = v2_add(up_pos, V2(-dx, 0));
		expand_down->pos = v2_add(down_pos, V2(-dx, 0));

		shrink_left->pos = v2_add(left_pos, V2(0, +dx));
		shrink_right->pos = v2_add(right_pos, V2(0, +dx));
		shrink_up->pos = v2_add(up_pos, V2(+dx, 0));
		shrink_down->pos = v2_add(down_pos, V2(+dx, 0));
	}
	{
		TextInput *start = &editor->starting_points_input, *inc = &editor->points_inc_input;
		start->pos = V2(width_gl-0.1f, 0.95f);
		inc->pos = V2(width_gl-0.1f, 0.9f);
		start->center = inc->center = true;
		start->width = inc->width = 0.15f;
		start->color = inc->color = points_color;
		start->char_predicate = inc->char_predicate = is_int_char;

		if (!start->focus && start->text[0] == '\0') strcpy(start->text, STRINGIFY(DFL_STARTING_POINTS));
		if (!inc->focus   && inc->text[0]   == '\0') strcpy(inc->text,   STRINGIFY(DFL_POINTS_INC));
	}

	{
		Event *e;
		for (e = input->events; e != events_end; ++e) {
			if (editor->dialog == EDIALOG_WARN_NEW && (e->kind == EVENT_KEY_DOWN || e->kind == EVENT_MOUSE_DOWN)) {
				// press any key or click to close the Ctrl+N unsaved changes warning
				editor->dialog = EDIALOG_NONE;
				continue; // skip this event
			}
			if (editor->dialog == EDIALOG_NONE) { // no dialog open; normal event processing 
				if (button_event(expand_left, state, e))
					resize_game(state, +1, 0, 0, 0);
				if (button_event(shrink_left, state, e))
					resize_game(state, -1, 0, 0, 0);
				if (button_event(expand_right, state, e))
					resize_game(state, 0, +1, 0, 0);
				if (button_event(shrink_right, state, e))
					resize_game(state, 0, -1, 0, 0);
				if (button_event(expand_down, state, e))
					resize_game(state, 0, 0, +1, 0);
				if (button_event(shrink_down, state, e))
					resize_game(state, 0, 0, -1, 0);
				if (button_event(expand_up, state, e))
					resize_game(state, 0, 0, 0, +1);
				if (button_event(shrink_up, state, e))
					resize_game(state, 0, 0, 0, -1);
				{
					TextInput *start = &editor->starting_points_input, *inc = &editor->points_inc_input;
					text_input_event(start, state, e);
					text_input_event(inc,   state, e);

					if (start->focus)
						editor->unsaved_changes = true;
					else
						text_input_validate_int(start, 0, STARTING_POINTS_MAX, 150);
					if (inc->focus)
						editor->unsaved_changes = true;
					else
						text_input_validate_int(inc, -STARTING_POINTS_MAX, +STARTING_POINTS_MAX, 25);
				}
				switch (e->kind) {
				case EVENT_KEY_DOWN: {
					KeyEvent *k = &e->u.key;
					switch (k->key) {
					case KEY_N:
						if (control_key_down) {
							if (shift_key_down || !editor->unsaved_changes) {
								editor_new(state);
							} else {
								// there are unsaved changes
								editor->dialog = EDIALOG_WARN_NEW;
							}
						}
						break;
					case KEY_O:
						if (control_key_down) {
							editor->dialog = EDIALOG_OPEN;
							text_input_clear(&editor->filename_input);
							editor->filename_input.focus = true;
						}
						break;
					case KEY_S:
						if (control_key_down) {
							if (shift_key_down)
								editor_save_as(state);
							else
								editor_save(state);
						}
						break;
					case KEY_T:
						// start placing tiles 
						editor->placing = ENT_TILE;
						editor->tile.kind = TL_REG;
						break;

					// start placing pions 
					case KEY_1: editor->placing = ENT_PION; editor->pion.player = 0; break;
					case KEY_2: editor->placing = ENT_PION; editor->pion.player = 1; break;
					}
				} break;
				case EVENT_MOUSE_DOWN: {
					MouseEvent *m = &e->u.mouse;
					uint button = m->button;
					if (!editor->dialog) editor_handle_mouse(state, m->x, m->y, 1U << button);
				} break;
				case EVENT_MOUSE_MOVE: {
					MouseMoveEvent *m = &e->u.mmove;
					if (m->mask)
						if (!editor->dialog)
							editor_handle_mouse(state, m->x, m->y, m->mask);
				} break;
				}
			}
		}
	}
	hex_setup_game_grid(state, grid_min_coords, grid_max_coords);

	{
		TextFont *font = big_font;
		font_begin(font);
		gl_color_text();
		{
			char str[MAP_NAME_BUFSZ+10] = {0};
			char const *name = *editor->map_name == '\0' ? "Untitled" : editor->map_name;
			if (editor->unsaved_changes)
				sprintf(str, "*%s*", name);
			else
				strcpy(str, name);
			text_render_centered(font, str, V2(center_x, 0.97f));
		}
		font_end();
	}
	{
		sprites_begin(state);
		// expand/shrink buttons 
		for (i = 0; i < 8; ++i)
			button_render(&arrow_buttons[i], state);
		sprites_end();
	}
	render_game_board(state, editor->dialog == EDIALOG_NONE, 1); // only highlight the mouse hex if no dialog is open
	{
		Rect bounds = game_grid_bounds(game);
		glBegin(GL_QUADS);
		gl_color2f(1.0f, 0.5f);
		rect_render_border(bounds, gl_pixel(state));
		glEnd();
	}
	{
		TextInput *start = &editor->starting_points_input, *inc = &editor->points_inc_input;
		TextFont *font = main_font;

		text_input_render(start, state);
		text_input_render(inc, state);

		font_begin(font);
		v4_gl_color(points_color);
		text_render_centered(font, "+", v2_sub(inc->pos, V2(inc->width * 0.5f + 0.02f, 0)));
		font_end();
	}

	if (!editor->dialog) { // ghosts 
		HexPos mouse_hex_pos = hex_at_pos(game, mouse_pos);
		if (hex_in_bounds(game, mouse_hex_pos)) {
			// draw ghost where mouse is 
			switch (editor->placing) {
			case ENT_NONE: break;
			case ENT_TILE:
				render_hexes_start(state);
				render_hex(state, hex_center_pos(game, mouse_hex_pos), game->hex_radius, V4(1,1,1,0.4f));
				render_hexes_end(state);
				break;
			case ENT_PION: {
				Pion p = editor_pion(editor, mouse_hex_pos);
				sprites_begin(state);
				render_pion(state, &p, V4(1,1,1,0.5f));
				sprites_end();
			} break;
			}
		}
	}
	
	if (editor->dialog) {
		EditorDialog dialog = editor->dialog;
		Rect dialog_rect = {0};
		float dialog_x1 = 0, dialog_y1 = 0, dialog_x2 = 0, dialog_y2 = 0;
		float dialog_width = 0.8f, dialog_height = 0.8f;
		char selected_map[MAP_NAME_BUFSZ] = {0};
		int map_status = -1;

		dialog_rect = rect_centered(V2(center_x, 0.5f), V2(dialog_width, dialog_height));
		rect_coords(dialog_rect, &dialog_x1, &dialog_y1, &dialog_x2, &dialog_y2);

		if (dialog == EDIALOG_WARN_OVERWRITE) {
			button_yes.text = "Yes";
			button_no.text = "No";
			button_yes.color = V4(0, 0.7f, 0, 1);
			button_no.color  = V4(0.7f, 0, 0, 1);
			button_yes.center = button_no.center = true;
			button_yes.pos = V2(dialog_x1 + 0.2f, 0.4f);
			button_no.pos = V2(dialog_x2 - 0.2f, 0.4f);
		}

		// dialog box 
		glBegin(GL_QUADS);
		glColor4f(0.05f,0.05f,0.08f,0.9f);
		{
			float x1 = dialog_x1, y1 = dialog_y1, x2 = dialog_x2, y2 = dialog_y2;
			glVertex2f(x1, y1);
			glVertex2f(x2, y1);
			glVertex2f(x2, y2);
			glVertex2f(x1, y2);

			gl_color1f(0.8f);
			rect_render_border(rect4(x1, y1, x2, y2), 1.0f/(float)state->win_height);
		}
		glEnd();

		{ // dialog title 
			TextFont *font = &state->main_font;
			char const *title = "";
			switch (dialog) {
			case EDIALOG_NONE: break;
			case EDIALOG_SAVE_AS: title = "Save as..."; break;
			case EDIALOG_OPEN: title = "Open..."; break;
			case EDIALOG_WARN_OVERWRITE: title = "Are you sure?"; break;
			case EDIALOG_WARN_NEW: title = "There are unsaved changes!"; break;
			}
			font_begin(font);
			gl_color_text();
			text_render_centered(font, title, V2(center_x, dialog_y2 - 0.025f));
			if (dialog == EDIALOG_WARN_OVERWRITE) {
				char str[MAP_NAME_BUFSZ + 64]={0};
				sprintf(str,  "Do you want to overwrite '%s'?", editor->overwritten_map_name);
				text_render_centered(font, str, V2(center_x, 0.6f));
				button_render(&button_yes, state);
				button_render(&button_no, state);
			} else if (dialog == EDIALOG_WARN_NEW) {
				text_render_centered(font, "Either save or press Ctrl+Shift+N to discard changes.", V2(center_x, 0.45f));
			}
			font_end();
		}

		if (dialog == EDIALOG_SAVE_AS || dialog == EDIALOG_OPEN) {
			TextInput *inp = &editor->filename_input;
			inp->placeholder = "Map name";
			inp->pos = V2(center_x, dialog_y2-0.08f);
			inp->center = true;
			inp->max_len = MAP_NAME_MAX_LEN;
			inp->char_predicate = can_char_be_used_in_map_filename;

			map_status = map_selector(frame, &editor->filename_input, dialog_y1, 0.75f, 0.75f, selected_map);
		}

		switch (dialog) {
		case EDIALOG_NONE: break;
		case EDIALOG_SAVE_AS: {
			if (map_status != -1) { // if we've chosen a map to save as
				assert(*selected_map);
				if (map_status == 0) {
					// user selected pre-existing map; warn about overwriting it
					editor->dialog = EDIALOG_WARN_OVERWRITE;
					strcpy(editor->overwritten_map_name, selected_map);
				} else {
					strcpy(editor->map_name, selected_map);
					editor_save(state);
					// close save as dialog when you save 
					editor->dialog = EDIALOG_NONE;
				}
			}
			break;
		} break;
		case EDIALOG_OPEN: {
			if (*selected_map) {
				if (map_status == 0) { // pre-existing map 
					editor_open(state, selected_map);
				} else {
					text_input_clear(&editor->filename_input);
				}
			}
		} break;
		case EDIALOG_WARN_OVERWRITE: {
			Event *e;
			for (e = events; e != events_end; ++e) {
				if (button_event(&button_yes, state, e)) {
					strcpy(editor->map_name, editor->overwritten_map_name);
					editor_save(state);
					editor->dialog = EDIALOG_NONE;
				}
				if (button_event(&button_no, state, e)) {
					editor->dialog = EDIALOG_NONE;
				}
			}
		} break;
		case EDIALOG_WARN_NEW: break;
		}
	} 
}

